#include "Illuminant.h"
#include "ColorUtils.h"

#include <algorithm>

using namespace FCG;

EPSG::Vec3<double> ColorUtils::xyzbar_to_xyY( const EPSG::Vec3<double>& xyzbar )
{
	double sum = xyzbar.x + xyzbar.y + xyzbar.z;

	double y = ( xyzbar.y < 0.001 ) ? 0.1 : 100 * xyzbar.y;
	return EPSG::Vec3<double>( xyzbar.x / sum, xyzbar.y / sum, y );
}

EPSG::Vec3<double> FCG::ColorUtils::xyzbar_to_XYZ( const EPSG::Vec3<double>& xyzbar )
{
	return xyY_to_XYZ( xyzbar_to_xyY( xyzbar ) );
}

EPSG::Vec3<double> FCG::ColorUtils::xyzbar_to_RGB( const EPSG::Vec3<double>& xyzbar )
{
	return XYZ_to_RGB( xyzbar_to_XYZ( xyzbar ) );
}

EPSG::Vec3<double> FCG::ColorUtils::xyY_to_XYZ( const EPSG::Vec3<double>& xyY )
{
	double Yy = xyY.z / xyY.y;

	return EPSG::Vec3<double>( Yy * xyY.x, xyY.z, Yy * (1 - xyY.x - xyY.y) );
}

EPSG::Vec3<double> FCG::ColorUtils::XYZ_to_RGB( const EPSG::Vec3<double>& XYZ )
{
	double M[3][3] = {	{  3.2404542, -1.5371385, -0.4985314 },
					    { -0.9692660,  1.8760108,  0.0415560 },
						{  0.0556434, -0.2040259,  1.0572252 }	};

	EPSG::Vec3<double> xyzW = Illuminant::instance()->getStandardIlluminant();
	double xn = XYZ.x / xyzW.x;
	double yn = XYZ.y / xyzW.y;
	double zn = XYZ.z / xyzW.z;

	double r = M[0][0] * xn + M[0][1] * yn + M[0][2] * zn;
	double g = M[1][0] * xn + M[1][1] * yn + M[1][2] * zn;
	double b = M[2][0] * xn + M[2][1] * yn + M[2][2] * zn;

	r = gammaCorrect( r );
	g = gammaCorrect( g );
	b = gammaCorrect( b );

	return EPSG::Vec3<double>( r, g, b );
}

EPSG::Vec3<double> FCG::ColorUtils::RGB_to_XYZ( const EPSG::Vec3<double>& RGB )
{
	double M[3][3] = {	{  0.4124564,  0.3575761,  0.1804375 },
						{  0.2126729,  0.7151522,  0.0721750 },
						{  0.0193339,  0.1191920,  0.9503041 }	};

	double var_R = RGB.x;        //R = From 0 to 255
	double var_G = RGB.y;        //G = From 0 to 255
	double var_B = RGB.z;        //B = From 0 to 255

	if ( var_R > 0.04045 ) var_R = pow( ( ( var_R + 0.055 ) / 1.055 ), 2.4 );
	else                   var_R = ( var_R / 12.92 );
	if ( var_G > 0.04045 ) var_G = pow( ( ( var_G + 0.055 ) / 1.055 ), 2.4 );
	else                   var_G = ( var_G / 12.92 );
	if ( var_B > 0.04045 ) var_B = pow( ( ( var_B + 0.055 ) / 1.055 ), 2.4 );
	else                   var_B = ( var_B / 12.92 );

	var_R *= 100;
	var_G *= 100;
	var_B *= 100;

	double x = ( M[0][0] * var_R + M[0][1] * var_G + M[0][2] * var_B );
	double y = ( M[1][0] * var_R + M[1][1] * var_G + M[1][2] * var_B );
	double z = ( M[2][0] * var_R + M[2][1] * var_G + M[2][2] * var_B );
	
	return EPSG::Vec3<double>( x, y, z );
}

double FCG::ColorUtils::gammaCorrect( double g )
{
	double vc = ( g > 0 ? g : -g );

	if ( vc > 0.0031308 ) 
		vc = 1.055 * pow( vc, 1.0/2.4 ) - 0.055;
	else                  
		vc = 12.92 * vc;

	return ( g > 0 ? vc : -vc );
}

EPSG::Vec3<double> FCG::ColorUtils::XYZ_to_Lab( const EPSG::Vec3<double>& XYZ )
{
	EPSG::Vec3<double> xyzW = Illuminant::instance()->getStandardIlluminant();
	double var_X = XYZ.x / xyzW.x;
	double var_Y = XYZ.y / xyzW.y;
	double var_Z = XYZ.z / xyzW.z;

	if ( var_X > 0.008856 ) var_X = pow( var_X, ( 1./3 ) );
	else                    var_X = ( 7.787 * var_X ) + ( 16.0/ 116 );
	if ( var_Y > 0.008856 ) var_Y = pow( var_Y, ( 1./3 ) );
	else                    var_Y = ( 7.787 * var_Y ) + ( 16.0/ 116 );
	if ( var_Z > 0.008856 ) var_Z = pow( var_Z, ( 1./3 ) );
	else                    var_Z = ( 7.787 * var_Z ) + ( 16.0/ 116 );

	return EPSG::Vec3<double>( ( 116 * var_Y ) - 16, 500 * ( var_X - var_Y ), 200 * ( var_Y - var_Z ) );
}